/*
 * Filename: CB_data_acquisition.cpp
 * Date: 3/25/11
 * Author: Alain Zarembowitch / MSS
 * Revision: 0.1
 * Development environment: Microsoft Visual C++ 2010 Express, free downloadable software (recommended over Eclipse/CDT/MinGW)
 *
 * Description: Example on how to control a data acquisition chain comprising a USB gateway (COM-5003-A),
 * a data acquisition module (COM-8002) and a RF module (COM-30XX).
 * This program uses SRT commands to set control registers in volatile memory (to alleviate the limitation on
 * the maximum number of write cycles with non-volatile control registers).
 *
 * This code is provided as an example. Additional defensive code writing statements should be added to detect
 * and report any abnormal events to the user. 
 * 
 * Requirement: requires ComBlock USB driver version 3 (on the ComBlock CD-ROM)
 * Reference document: http://msdn.microsoft.com/en-us/library/ff540174%28v=vs.85%29.aspx#filehandle
 * Other helpful links: http://www.microsoft.com/whdc/connect/usb/default.mspx
 */

#include "main.h"
int CB_gap_ms = 10;	// gap in ms after sending a command for which no command is expected. 
					// Adjust as needed. Too small a delay may lead to unreliable controls
int retval = 0;
char buf[512];		// This buffer must be a multiple of the maximum packet size (64)
unsigned char regval;
int deviceIndex = 0;	// index for the USB device we want to communicate with (must be in the range [0,nDevices-1])
unsigned long upload_start_addr = 0; // see COM-8002 specifications
unsigned long upload_window_length = 0; // see COM-8002 specifications
unsigned char reg16 = 0x20;	// see COM-8002 specifications
unsigned char reg17 = 0x20;	// see COM-8002 specifications
unsigned char reg18 = 0x00;	// gain control value, unused when the AGC is enabled. see COM-8002 specifications
unsigned long input_decimation = 1;
unsigned long download_start_addr = 0; // see COM-8002 specifications
unsigned long download_window_length = 10000000; // see COM-8002 specifications

int _tmain(int argc, _TCHAR* argv[])
{
	char databuf[4096];	// larger buffer used for high-speed data transfer
	long nDevices;	// number of devices matching the device interface ID currently accessible

	// how many ComBlocks are currently reacheable via USB?
	// There could be more than one USB connection between ComBlocks and this PC
	// returns a negative number if error
	nDevices = comblock_USB_nActiveDevices(3);	// check only devices compatible with USB software rev3

    printf("Found %d USB connection(s) with ComBlock(s) \r\n", nDevices);
	if (nDevices <= 0){
		// no accessible device
		printf("Are the ComBlocks powered?\r\n", nDevices);
		system("PAUSE");
		return 0;
	}

	// Get a device handle for the device at the specified index deviceIndex
	retval = comblock_USB_openHandle(deviceIndex);
    if (retval < 0) 
    {
        printf("Could not get a device handle for the specified index %d  errNo=%d\r\n", deviceIndex, retval);
		return 0;
    }

	// ComBlock assembly initialization: Assign IDs after power up
	if(comblock_init()){
		return 0;
	}

	// COM-5003 USB gateway initialization
	retval = com5003_init();
	if(retval < 0)
		goto done;

	// COM-8002 data acquisition module: get ready for data acquisition (i.e. "upload")
	// upload address in bytes, upload window length in bytes, reg16, reg17, input decimation
	upload_start_addr = 0 & 0xFFFFFFC0;	// store data above base address 0, always a multiple of 32 Bytes
	upload_window_length = 10000000 & 0xFFFFFFC0; // capture 10MB, always a multiple of 32 Bytes
	retval = com8002_upload_init();
	
	// COM-8002 data acquisition module: start upload
	retval = com8002_start_upload();

	// check COM-8002 upload progress
	while(1){
		Sleep(100);	// query write pointer repetitively every 100ms
		// read write pointer address
		retval = com8002_get_wptr();
		if(retval >= upload_start_addr + upload_window_length)
			break;	// done uploading
	}
	
	// upload complete. 
	com8002_stop_upload();
	printf("Upload is complete \r\n");
	system("PAUSE");

	// COM-8002 data acquisition module: get ready for data download to a file
	download_start_addr = upload_start_addr;
	download_window_length = upload_window_length;
	retval = com8002_download_init();

	// open destination file
	FILE * pFile;
	pFile = fopen ("output.bin","wb");	// write, binary

	// read and discard any data which could be stuck in USB elastic buffer
	for(int i = 0; i < 8; i++){	// read enough times to empty a 16Kbit elastic buffer in the FPGA.
		retval = comblock_USB_Read(2, sizeof(databuf), databuf);
		if((retval == 0) || (retval == -19))
			break;	// nothing to read.
	}
	if(retval > 0){
		// abnormal condition. why are we still receiving data? maybe COM-5003 is in test throughput mode?
		printf("Abnormal condition: is the COM-5003 in throughput test mode??? \r\n");
		goto done2;
	}

	// tell the COM-8002 to start downloading data
	retval = com8002_start_download();

	// read data from the USB data stream 2 (high-speed data), as fast as we can and save to file.
	// this section could be optimized for speed as it is inefficient to write small buffers to the hard disk drive.
	unsigned long totalRxBytes;
	totalRxBytes = 0;
	int databufsize = sizeof(databuf);
	while(1){
		// adjust the buffer size to read the last few remaining bytes 
		if((download_window_length - totalRxBytes) < sizeof(databuf)){
			databufsize = download_window_length - totalRxBytes;
		}
		retval = comblock_USB_Read(2, databufsize, databuf);
		if(retval > 0){
			totalRxBytes += retval;	// total number of bytes received so far
		    // printf("%d\r\n", totalRxBytes);	// display progress
			// write to file
			fwrite(databuf, 1, retval, pFile);
		} else if(retval == -19){
			// timeout
			Sleep(100);
		}
		if(totalRxBytes >= download_window_length){
			printf("Download is complete \r\n");
			break;	// download complete
		}
	}
done2:
	// stop download
	com8002_stop_download();
	// close output file
	fclose(pFile);

done:
    comblock_USB_closeHandle();
	system("PAUSE");
	return 0;
}

//--------------------------------------------------------------------------------
// Enumerate ComBlocks after power-up
// /*Also reset the COM-5003 to clear stuck bytes in elastic buffer*/
// This is equivalent to clicking on the 2nd button to the left in the ComBlock Control Center
// graphical user interface.
int comblock_init(void){
	/*
	// ComBlock assembly initialization: RESET
	sprintf(buf,"@000RST\r\n");	// reset 1st board (COM-5003). Reason: to clear any data stuck in elastic buffers.
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	// repeat reset for address 001 because the COM-5003 may have been assigned an address once before.
	sprintf(buf,"@001RST\r\n");	// reset 1st board (COM-5003). Reason: to clear any data stuck in elastic buffers.
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
    comblock_USB_closeHandle();	// release the USB handle since the target is reset.
	Sleep(3000); // wait 3s for the COM-5003 to boot-up after a reset.
	// Get a device handle (again) for the device at the specified index deviceIndex
	retval = comblock_USB_openHandle(deviceIndex);
    if (retval < 0) 
    {
        printf("Could not get a device handle for the specified index %d  errNo=%d\r\n", deviceIndex, retval);
		return -1;
    }
	*/

	// ComBlock assembly initialization: Assign IDs after power up
	sprintf(buf,"@000SAC001\r\n");	// assign ComBlock ID 001 to first ComBlock from the USB connection (COM-5003)	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	sprintf(buf,"@001MFW9\r\n");	// forward messages to neighbouring ComBlocks
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	sprintf(buf,"@000SAC002\r\n");	// assign ComBlock ID 002 to second ComBlock from the USB connection (COM-8002)
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	sprintf(buf,"@002MFW9\r\n");	// forward messages to neighbouring ComBlocks	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	sprintf(buf,"@000SAC003\r\n");	// assign ComBlock ID 003 to second ComBlock from the USB connection (COM-3010)
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	sprintf(buf,"@003MFW9\r\n");	// forward messages to neighbouring ComBlocks	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	return 0;
}

//--------------------------------------------------------------------------------
// COM-5003 USB gateway initialization
int com5003_init(void){

	// check that the correct option (-A for USB) is running
	// first of all, make sure no data is stuck in rx pipe. Read until rx pipe is empty.
	while(1){
		retval = comblock_USB_Read(1, sizeof(buf), &buf[0]);
		if(retval < 0)
			break;
	}
	sprintf(buf,"@001GMI\r\n");
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// reply should be ready after delay (ComBlock micro is a bit slow)
	retval = comblock_USB_Read(1, sizeof(buf), &buf[0]);
	if(retval <= 0){
		return 0; // no reply
	}
	// check if the reply is complete
	retval = is_answer_complete(buf, retval);
	if(retval <= 0){
		return 0; // incomplete reply
	}
	// analyze message
	if(!memcmp(buf,"@999MID5003B",12)){
		// error. wrong COM-5003 option. Must be -A to work with high-speed USB 
		printf("Wrong COM-5003 option. Must be -A to work with high-speed USB\r\nPlease switch to personality index 2.");
		return -1;
	}

	// configure COM-5003 input for 8-bit parallel input, no test mode
	regval = 0x08;
	sprintf(buf,"@001SRT20%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	return 0;
}

//--------------------------------------------------------------------------------
// COM-8002 data acquisition module: get ready for data acquisition (i.e."upload")
int com8002_upload_init(void){

	// upload start address in bytes
	regval = (unsigned char)(upload_start_addr & 0x000000FF);
	sprintf(buf,"@002SRT00%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((upload_start_addr >> 8) & 0x000000FF);
	sprintf(buf,"@002SRT01%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((upload_start_addr >> 16) & 0x000000FF);
	sprintf(buf,"@002SRT02%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((upload_start_addr >> 24) & 0x000000FF);
	sprintf(buf,"@002SRT03%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// upload window length in bytes
	regval = (unsigned char)(upload_window_length & 0x000000FF);
	sprintf(buf,"@002SRT04%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((upload_window_length >> 8) & 0x000000FF);
	sprintf(buf,"@002SRT05%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((upload_window_length >> 16) & 0x000000FF);
	sprintf(buf,"@002SRT06%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((upload_window_length >> 24) & 0x000000FF);
	sprintf(buf,"@002SRT07%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// REG16: AGC enabled, no action, external trigger disabled. 
	sprintf(buf,"@002SRT16%02X\r\n",reg16);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// REG17: input format: 20-bit wide
	sprintf(buf,"@002SRT17%02X\r\n",reg17);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// input decimation 
	regval = (unsigned char)(input_decimation & 0x000000FF);
	sprintf(buf,"@002SRT19%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((input_decimation >> 8) & 0x000000FF);
	sprintf(buf,"@002SRT20%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((input_decimation >> 16) & 0x000000FF);
	sprintf(buf,"@002SRT21%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// REG18: gain control. Write to REG18 to enact the previous register changes
	sprintf(buf,"@002SRT18%02X\r\n",reg18);	// send command
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	return 0;
}

//--------------------------------------------------------------------------------
// COM-8002 data acquisition module: start upload
int com8002_start_upload(void){

	// REG16: start upload
	reg16 = (reg16 & 0xF1) | 0x02;
	sprintf(buf,"@002SRT16%02X\r\n",reg16);
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// REG18: gain control. Write to REG18 to enact the previous register changes
	sprintf(buf,"@002SRT18%02X\r\n",reg18);	// send command
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	return 0;
}

//--------------------------------------------------------------------------------
// COM-8002 data acquisition module: stop upload
int com8002_stop_upload(void){

	// REG16: stop upload
	reg16 = (reg16 & 0xF1) | 0x04;
	sprintf(buf,"@002SRT16%02X\r\n",reg16);
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// REG18: gain control. Write to REG18 to enact the previous register changes
	sprintf(buf,"@002SRT18%02X\r\n",reg18);	// send command
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	return 0;
}

//--------------------------------------------------------------------------------
// COM-8002 data acquisition module: get ready for data download
int com8002_download_init(void){

	// download start address in bytes
	regval = (unsigned char)(download_start_addr & 0x000000FF);
	sprintf(buf,"@002SRT08%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((download_start_addr >> 8) & 0x000000FF);
	sprintf(buf,"@002SRT09%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((download_start_addr >> 16) & 0x000000FF);
	sprintf(buf,"@002SRT10%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((download_start_addr >> 24) & 0x000000FF);
	sprintf(buf,"@002SRT11%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// download window length in bytes
	regval = (unsigned char)(download_window_length & 0x000000FF);
	sprintf(buf,"@002SRT12%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((download_window_length >> 8) & 0x000000FF);
	sprintf(buf,"@002SRT13%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((download_window_length >> 16) & 0x000000FF);
	sprintf(buf,"@002SRT14%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	regval = (unsigned char)((download_window_length >> 24) & 0x000000FF);
	sprintf(buf,"@002SRT15%02X\r\n",regval);	// send command	
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// REG18: gain control. Write to REG18 to enact the previous register changes
	sprintf(buf,"@002SRT18%02X\r\n",reg18);	// send command
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	return 0;
}

//--------------------------------------------------------------------------------
// COM-8002 data acquisition module: start download
int com8002_start_download(void){

	// REG16: start download
	reg16 = (reg16 & 0xF1) | 0x06;
	sprintf(buf,"@002SRT16%02X\r\n",reg16);
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)

	// REG18: gain control. Write to REG18 to enact the previous register changes
	sprintf(buf,"@002SRT18%02X\r\n",reg18);	// send command
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	return 0;
}

//--------------------------------------------------------------------------------
// COM-8002 data acquisition module: stop download
int com8002_stop_download(void){

	// REG16: stop download
	reg16 = (reg16 & 0xF1) | 0x08;
	sprintf(buf,"@002SRT16%02X\r\n",reg16);
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slower than PC)

	// REG18: gain control. Write to REG18 to enact the previous register changes
	sprintf(buf,"@002SRT18%02X\r\n",reg18);	// send command
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// wait CB_gap_ms between successive commands (ComBlock micro is slow)
	return 0;
}

//--------------------------------------------------------------------------------
// COM-8002 data acquisition module: read write pointer location as a way to 
// check on the upload progress. Reads status registers SREG19,20,21,22 in sequence.
// Returns the write pointer location in bytes. 
unsigned long com8002_get_wptr(void){
	unsigned long wptr = 0;
	unsigned long a = 0;

	for(int sregno = 19; sregno<= 22; sregno++){
		a = GSR(2, sregno);
		// shift in status register bytes from left to right, LSB first.
		wptr = (wptr >> 8) | ((a << 24)& 0xFF000000);
	}

	return wptr;
}

//--------------------------------------------------------------------------------
// get status register from designated Comblock module. Returns byte.
unsigned char GSR(int moduleID, int sregno){
	unsigned char a = 0;

	// first of all, make sure no data is stuck in rx pipe. Read until rx pipe is empty.
	while(1){
		retval = comblock_USB_Read(1, sizeof(buf), &buf[0]);
		if(retval < 0)
			break;
	}
	sprintf(buf,"@%03dGSR%02d\r\n",moduleID, sregno);
	retval = comblock_USB_Write(1,strlen(buf), buf);	// send command to ComBlock
	Sleep(CB_gap_ms);	// reply should be ready after delay (ComBlock micro is a bit slow)
	retval = comblock_USB_Read(1, sizeof(buf), &buf[0]);
	if(retval <= 0){
		return 0; // no reply
	}
	// check if the reply is complete
	retval = is_answer_complete(buf, retval);
	if(retval <= 0){
		return 0; // incomplete reply
	}
	// analyze message
	buf[9] = 0;	// create end of string after the 2 characters of interest.
	sscanf(&buf[7],"%2X", &regval);
	return regval;
}


//--------------------------------------------------------------------------------
// checks if the string in buf is a complete message. i.e. should start with @ and end with \n
// retuns string length and re-formatted message in buf.
int is_answer_complete(char * buf, int length){
	char *p1, *p2;
	// find location of first '@' character in buf
	p1 = strchr(buf, '@');
	if(p1 == NULL){
		// no such character. ignore entire buffer
		return -1;
	}

	// find location of first '\n' character in buf
	p2 = strchr(buf, '\n');
	if(p2 == NULL){
		// no such character. ignore entire buffer. Reply is incomplete
		return -2;
	}

	// reformat buf as a string starting with @, ending with \n and 0 at the end of string
	if(p1 != &buf[0]){
		memcpy(buf,p1,p2-p1+1);
	}
	// add 0 at the end to terminate the string
	buf[p2-p1+1] = 0;
	return (p2-p1+1);
}


//--------------------------------------------------------------------------------




