/*
 * Filename: WinUSBsimpleApp.cpp
 * Date: 12/17/10
 * Author: Alain Zarembowitch / MSS
 * Revision: 0.1
 * Development environment: Microsoft Visual C++ 2010 Express, free downloadable software (recommended over Eclipse/CDT/MinGW)
 *
 * Description: Example to write a command (@000GMI) to a Comblock connected over USB, then
 * read the ComBlock reply (@999MID...).
 * Requirement: requires ComBlock USB driver version 3 (on the ComBlock CD-ROM)
 * Reference document: http://msdn.microsoft.com/en-us/library/ff540174%28v=vs.85%29.aspx#filehandle
 * Other helpful links: http://www.microsoft.com/whdc/connect/usb/default.mspx
 */

#include "main.h"

int _tmain(int argc, _TCHAR* argv[])
{
	// Constant for ComBlock USB device interface
	// (same number as in the driver .inf file, rev3.0.0.1)
	const GUID COMBLOCK_USB_DEVICE_INTERFACE = { 0x4416e0fd, 0x9efc, 0x4b5b, { 0x8d, 0xe8, 0x31, 0x79, 0x8a, 0x45, 0x81, 0x39 } };
	// (same number as in the driver .inf file, rev2.0.0.1)
	//static const GUID COMBLOCK_USB_DEVICE_INTERFACE = { 0x4D36E97E, 0xE325, 0x11CE, { 0xBF, 0xC1, 0x08, 0x00, 0x2B, 0xE1, 0x03, 0x18 } };

	HANDLE hDeviceHandle = NULL; 
    BOOL bResult = TRUE;
	WINUSB_INTERFACE_HANDLE hWinUSBHandle; 
	unsigned long errNo = 0;
    USB_INTERFACE_DESCRIPTOR InterfaceDescriptor;
	unsigned char bNumEndpoints = 0;	// number of endpoints in the interface descriptor (2 or 4?)
	struct PIPE_ID
	{
		UCHAR  PipeInId;
		UCHAR  PipeOutId;
	};
    WINUSB_PIPE_INFORMATION  Pipe[4];	// ComBlocks support at most 4 pipes (2 for monitoring & control, 2 for high-speed data xfers)
	char buf[512];		// The buffer must be a multiple of the maximum packet size (64)
	unsigned long l;
	unsigned long bytes_sent = 0;
	unsigned long bytes_read = 0;
	OVERLAPPED Overlapped;	// overlapped structure
	HANDLE hEvent;	// event for overlapped operation
	unsigned long check = 0;
	long mc_in_index;
	long mc_out_index;
	long nDevices;	// number of devices matching the device interface ID currently accessible
	long deviceIndex = 0;	// index for the USB device we want to communicate with (must be in the range [0,nDevices-1])

	// how many ComBlocks are currently reacheable via USB?
	// There could be more than one USB connection between ComBlocks and this PC
	// returns a negative number if error
	nDevices = nActiveDevices(COMBLOCK_USB_DEVICE_INTERFACE);

    printf("Found %d USB connection(s) with ComBlock(s) \r\n", nDevices);
	if (nDevices <= 0){
		// no accessible device
		system("PAUSE");
		return 0;
	}

	// Get a device handle for the device at the specified index deviceIndex
	bResult = GetDeviceHandle(COMBLOCK_USB_DEVICE_INTERFACE, deviceIndex, &hDeviceHandle);
    if (!bResult) 
    {
        errNo = GetLastError();
        printf("Could not get a device handle for the specified index %d  errNo=%d\r\n", deviceIndex, errNo);
		return 0;
    }

	// Get a WinUSB Interface Handle for the Device
    bResult = WinUsb_Initialize(hDeviceHandle, &hWinUSBHandle);
    if (!bResult) 
    {
        errNo = GetLastError();
		if(errNo == ERROR_INVALID_HANDLE){
			printf("Invalid handle\r\n");
		} else if(errNo == ERROR_NOT_ENOUGH_MEMORY){
			printf("Insufficient memory to do the operation\r\n");
		} else if(errNo == ERROR_BAD_DEVICE){
			printf("Default device descriptor not found\r\n");
		} else {
	        printf("Could not get a WinUSB interface handle for the specified index %d  errNo=%d\r\n", deviceIndex, errNo);
		}
	    CloseHandle(hDeviceHandle);
		return 0;
    }

	// Get Interface Descriptors from the device itself
	// One interface
	memset(&InterfaceDescriptor, 0, sizeof(USB_INTERFACE_DESCRIPTOR));	// fill in with zeros
    bResult = WinUsb_QueryInterfaceSettings(hWinUSBHandle, 0, &InterfaceDescriptor);
	if(!bResult)
    {
        errNo = GetLastError();
		printf("Error getting interface descriptor %d.\r\n", errNo);
        goto done;
    }
	bNumEndpoints = InterfaceDescriptor.bNumEndpoints;	// should be 2 (M&C only) or 4 (M&C and Data)
	printf("Interface descriptor reports %d endpoints \r\n",bNumEndpoints);

	// Get Endpoints Descriptors from the device itself (2 or 4 endpoints)
	// Find ComBlock M&C IN and OUT ports
	mc_in_index = -1;	// yet unknown ComBlock monitoring and control IN port
	mc_out_index = -1;	// yet unknown ComBlock monitoring and control OUT port
    for (int index = 0; index < bNumEndpoints; index++)
    {
		memset(&Pipe[index],0,sizeof(WINUSB_PIPE_INFORMATION)); // Fills a block of memory with zeros.
        bResult = WinUsb_QueryPipe(hWinUSBHandle, 0, index, &Pipe[index]);
		if(!bResult)
		{
			errNo = GetLastError();
			printf("Error %d getting endpoint descriptor %d.\r\n", errNo, index);
	        goto done;
		}
		printf("    Endpoint %d: ID=%02X Type=%02X\r\n",index+1,Pipe[index].PipeId,Pipe[index].PipeType);
		if((mc_in_index < 0) && (Pipe[index].PipeId & 0x80)){
			// first IN port is the one for ComBlock monitoring and control
			mc_in_index = index;
		} else if((mc_out_index < 0) && !(Pipe[index].PipeId & 0x80)){
			// first OUT port is the one for ComBlock monitoring and control
			mc_out_index = index;
		}
	}

	// Send Control Transfer to the Default Endpoint
	// N/A here. Nothing to do.

	// Issue I/O Requests
	// But first, create an event object  
    hEvent = CreateEvent( 
        NULL,    // default security attribute 
        TRUE,    // manual-reset event 
        TRUE,    // initial state = signaled 
        NULL);   // unnamed event object 

      if (hEvent == NULL) 
      {
		printf("CreateEvent failed with %d.\n", GetLastError()); 
        goto done;
      }


	// Example: ask ComBlock for its identification (@000GMI\r\n command, see ComBlock API)
	sprintf(buf,"@000GMI\r\n");
	l = strlen(buf);
	// make sure message does not exceed destination limit.
	if(l > MAX_MESSAGE_LENGTH)
		l = MAX_MESSAGE_LENGTH;
	// non-blocking call
	Overlapped.hEvent = hEvent;
	bResult = WinUsb_WritePipe(hWinUSBHandle, Pipe[mc_out_index].PipeId, (unsigned char *)(buf), l, &bytes_sent, &Overlapped);
    if(!bResult)
    {
		errNo = GetLastError();
		if(errNo == ERROR_IO_PENDING){
			// write operating still pending
			check = 0;
			while(1){
				// non blocking call. returns immediately.
				bytes_sent = 0;
				bResult = WinUsb_GetOverlappedResult(hWinUSBHandle, &Overlapped, &bytes_sent, 0);
				check++;
				if(check == 100){
					// time out
					printf("Timeout writing to USB device. Error %d.\r\n", errNo);
					goto done;
				}
				if ((!bResult) || (bytes_sent == 0)){
					Sleep(1);
					continue;
				}
				if(bytes_sent == l)
					break;
			}
		} else {
			printf("Error %d writing to ComBlock.\r\n", errNo);
	        goto done;
		}
    }

	// request a read from ComBlock. Asynchronous. returns immediately. 
	Overlapped.hEvent = hEvent;
	memset(buf,0,sizeof(buf));	// clear buffer
	bResult = WinUsb_ReadPipe(hWinUSBHandle, Pipe[mc_in_index].PipeId, (unsigned char *)(buf), sizeof(buf), &bytes_read, &Overlapped);
    if(!bResult)
    {
		errNo = GetLastError();
		if(errNo == ERROR_IO_PENDING){
			// read operating still pending
			check = 0;
			while(1){
				// non blocking call. returns immediately.
				bytes_read = 0;
				bResult = WinUsb_GetOverlappedResult(hWinUSBHandle, &Overlapped, &bytes_read, 0);
				check++;
				if(check == 100){
					// time out
					printf("Timeout reading from USB device. Error %d.\r\n", errNo);
					goto done;
				}
				if ((!bResult) || (bytes_read == 0)){
					Sleep(1);
					continue;
				}
				// check if the ComBlock reply is complete (should start with @ and end with \n)
				if(is_answer_complete(buf,bytes_read))
					break;
			}
		} else {
			printf("Error %d reading from ComBlock.\r\n", errNo);
			goto done;
		}
    }

	printf("read %d bytes %s\r\n",bytes_read, buf);
	system("PAUSE");
	return 0;

done:
    CloseHandle(hDeviceHandle);
    WinUsb_Free(hWinUSBHandle);
	system("PAUSE");

	return 0;
}

//--------------------------------------------------------------------------------
// checks if the string in buf is a complete message. i.e. should start with @ and end with \n
// retuns string length and re-formatted message in buf.
int is_answer_complete(char * buf, int length){
	char *p1, *p2;
	// find location of first '@' character in buf
	p1 = strchr(buf, '@');
	if(p1 == NULL){
		// no such character. ignore entire buffer
		return -1;
	}

	// find location of first '\n' character in buf
	p2 = strchr(buf, '\n');
	if(p2 == NULL){
		// no such character. ignore entire buffer. Reply is incomplete
		return -2;
	}

	// reformat buf as a string starting with @, ending with \n and 0 at the end of string
	if(p1 != &buf[0]){
		memcpy(buf,p1,p2-p1+1);
	}
	// add 0 at the end to terminate the string
	buf[p2-p1+1] = 0;
	return (p2-p1+1);
}


//--------------------------------------------------------------------------------
// checks how many active devices match the device interface ID
int nActiveDevices(GUID guidDeviceInterface){
    HDEVINFO hDeviceInfo;
    SP_DEVINFO_DATA DeviceInfoData;
	int i; 
    BOOL bResult;
	long errNo = 0;

	if (guidDeviceInterface==GUID_NULL)
    {
        return 0;
    }
    // Get information about all the installed devices for the specified
    // device interface class.
    hDeviceInfo = SetupDiGetClassDevs( 
        &guidDeviceInterface,
        NULL, 
        NULL,
        DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

    if (hDeviceInfo == INVALID_HANDLE_VALUE) 
    { 
		errNo = GetLastError();	// Error SetupDiGetClassDevs
		return -1;	
    }

    //Enumerate all the device interfaces in the device information set.
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);
	for(i = 0; ; i++){
		bResult = SetupDiEnumDeviceInfo(hDeviceInfo, i, &DeviceInfoData);
		if(!bResult)
			break;
	}
	return i;

}

//--------------------------------------------------------------------------------
// Get a device handle for the device at the specified index
BOOL GetDeviceHandle (GUID guidDeviceInterface, int deviceIndex, PHANDLE hDeviceHandle){
    BOOL bResult = TRUE;
    HDEVINFO hDeviceInfo;
    SP_DEVINFO_DATA DeviceInfoData;
    SP_DEVICE_INTERFACE_DATA deviceInterfaceData;
    PSP_DEVICE_INTERFACE_DETAIL_DATA pInterfaceDetailData = NULL;
    ULONG requiredLength=0;
    LPTSTR lpDevicePath = NULL;

    if (guidDeviceInterface==GUID_NULL)
    {
        return FALSE;
    }

    // Get information about all the installed devices for the specified
    // device interface class.
    hDeviceInfo = SetupDiGetClassDevs( 
        &guidDeviceInterface,
        NULL, 
        NULL,
        DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);

    if (hDeviceInfo == INVALID_HANDLE_VALUE) 
    { 
        // ERROR 
        printf("Error SetupDiGetClassDevs: %d.\n", GetLastError());
        goto done;
    }

    //Enumerate all the device interfaces in the device information set.
    DeviceInfoData.cbSize = sizeof(SP_DEVINFO_DATA);

	SetupDiEnumDeviceInfo(hDeviceInfo, deviceIndex, &DeviceInfoData);
	// The SetupDiEnumDeviceInfo function returns a SP_DEVINFO_DATA structure that specifies 
	// a device information element in a device information set

	deviceInterfaceData.cbSize = sizeof(SP_INTERFACE_DEVICE_DATA);

    //Get information about the device interface.
    bResult = SetupDiEnumDeviceInterfaces( 
        hDeviceInfo,
        &DeviceInfoData,
        &guidDeviceInterface,
        0, 
        &deviceInterfaceData);

    //Check for some other error
    if (!bResult) 
    {
        printf("Error SetupDiEnumDeviceInterfaces: %d.\n", GetLastError());
        goto done;
    }

    //Interface data is returned in SP_DEVICE_INTERFACE_DETAIL_DATA
    //which we need to allocate, so we have to call this function twice.
    //First to get the size so that we know how much to allocate
    //Second, the actual call with the allocated buffer
        
    bResult = SetupDiGetDeviceInterfaceDetail(
        hDeviceInfo,
        &deviceInterfaceData,
        NULL, 0,
        &requiredLength,
        NULL);


    //Check for some other error
    if (!bResult) 
    {
        if ((ERROR_INSUFFICIENT_BUFFER==GetLastError()) && (requiredLength>0))
        {
            //we got the size, allocate buffer
            pInterfaceDetailData = (PSP_DEVICE_INTERFACE_DETAIL_DATA)LocalAlloc(LPTR, requiredLength);
                
            if (!pInterfaceDetailData) 
            { 
                // ERROR 
                printf("Error allocating memory for the device detail buffer.\n");
                goto done;
            }
        }
        else
        {
            printf("Error SetupDiEnumDeviceInterfaces: %d.\n", GetLastError());
            goto done;
        }
    }

    //get the interface detailed data
    pInterfaceDetailData->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

    //Now call it with the correct size and allocated buffer
    bResult = SetupDiGetDeviceInterfaceDetail(
            hDeviceInfo,
            &deviceInterfaceData,
            pInterfaceDetailData,
            requiredLength,
            NULL,
            &DeviceInfoData);
        
    //Check for some other error
    if (!bResult) 
    {
        printf("Error SetupDiGetDeviceInterfaceDetail: %d.\n", GetLastError());
        goto done;
    }

    //copy device path
                
    size_t nLength = wcslen (pInterfaceDetailData->DevicePath) + 1;  
    lpDevicePath = (TCHAR *) LocalAlloc (LPTR, nLength * sizeof(TCHAR));
    StringCchCopy(lpDevicePath, nLength, pInterfaceDetailData->DevicePath);
    lpDevicePath[nLength-1] = 0;	// string termination
                        
    printf("Device path:  %s\n", lpDevicePath);


    if (!lpDevicePath)
    {
        //Error.
        printf("Error %d.", GetLastError());
        goto done;
    }

    //Open the device
    *hDeviceHandle = CreateFile (
        lpDevicePath,
        GENERIC_READ | GENERIC_WRITE,
        FILE_SHARE_READ | FILE_SHARE_WRITE,
        NULL,
        OPEN_EXISTING,
        FILE_FLAG_OVERLAPPED,
        NULL);

    if (*hDeviceHandle == INVALID_HANDLE_VALUE)
    {
        //Error.
        printf("Error %d.", GetLastError());
        goto done;
    }



done:
    LocalFree(lpDevicePath);
    LocalFree(pInterfaceDetailData);    
    bResult = SetupDiDestroyDeviceInfoList(hDeviceInfo);
    
    return bResult;
}

//--------------------------------------------------------------------------------




