// USB.java
// Java interface for USB Application
package comblock;

/**
 * <p>Title: USB </p>
 * <p>Description: Java Interface for USB Application</p>
 * <p>Copyright: Copyright (c) 2005-2010</p>
 * <p>Company: MSS</p>
 * @author Poornima Sarvadevabhatla / MSS
 * @author Alain Zarembowitch / MSS
 * @version 3.0
 * 
 * Reference documents: 
 * - http://www.rgagnon.com/javadetails/java-0348.html
 * - http://www.cs.umanitoba.ca/~eclipse/8-JNI.pdf
 * 
 * 
 * Rev 3 1/9/11 AZ
 * Stop using the separate usb.jar 
 * Change driver to winusb for compatibility with newer Windows operating systems (7)
 * Use usbdll.dll (instead of the previous rev usbcpp.dll)
 * Removed exceptions. Errors are returned as negative numbers.
 * Made the native methods public... no need for another wrapper.
 * Added method to empty rx queue
 */
import java.util.Timer;
import java.util.TimerTask;

public class USB {
	// Load the dll
	private boolean timeIsUp = false;
	private TimerTask timeoutTimerTask = null;
	static {
		System.loadLibrary("usbdll"); // rev3 usb driver  (previously usbcpp in rev2)
	}

	//--------------------------------------------------------------------------------
	//Method: nActiveDevices
	//This method detects how many ComBlocks are connected via USB and ready to communicate.
	//driver_rev_no can be 2(now obsolete version. Just checking if the user need to install a new driver) 
	//or 3 (current version).
	//
	//Returns the number of ComBlocks directly connected to this PC via USB.
	//
	//Error codes: 
	//-1 invalid handle value
	//-2 invalid driver rev number
	//--------------------------------------------------------------------------------
	public native int nActiveDevices(int driver_rev_no);

	//--------------------------------------------------------------------------------
	//Method: openHandle
	//Get a device handle for the device at the specified index
	//The handle will be kept for all other methods until the close method is called.
	//Uses usb device driver rev3 (GUID) only (ignores previous driver rev)
	//The java application can only communicate with one driver at a time.
	//Attempt to open multiple handles will result in closing previous ones.
	//
	//This method does more than just creating a device handle and a winusb handle, 
	//it also scans the endpoint descriptor to find the correct endpoint locations for
	//the two monitoring and control pipes.
	//
	//Returns: 0 when everything went OK.
	//
	//Error codes: 
	//-1 device index is out of range. Valid range [0, nDevices-1]
	//-2 error getting a device handle
	//-3 Could not get a WinUSB interface handle for the specified index
	//-4 Error getting interface descriptor
	//-5 Error getting endpoint descriptor
	//-6 could not find the two M&C endpoint addresses in the endpoint descriptors
	//--------------------------------------------------------------------------------
	public native int openHandle(int device_index);

	//--------------------------------------------------------------------------------
	//Method: closeHandle
	//Close the existing handles (file + winusb)
	//Returns: n/a
	//
	//Error codes: n/a
	//--------------------------------------------------------------------------------
	public native void closeHandle();

	//--------------------------------------------------------------------------------
	//Method: Write
	//
	//Write to data stream 1 (M&C) or 2 (high-speed data)
	//
	//Returns the number of bytes actually sent (could be less if there are communication problems)
	// Blocking until either data is sent or 100ms timeout.
	//
	//Error codes: 
	//-10 no handle to the driver is present. User needs to call openHandle() first.
	//-11 Missing a pointer to the buffer holding the data to be written
	//-13 Can't write a negative number of bytes
	//-14 Could not detect M&C OUT endpoint in descriptors
	//-15 Could not detect high-speed data OUT endpoint in descriptors
	//-16 invalid dataStream input selection
	//-18 Some other error writing to USB device
	//-19 timeout
	//--------------------------------------------------------------------------------
	public native int Write(int dataStream, int noOfBytesToWrite,
			byte[] bufferWithValuesToWrite);

	//--------------------------------------------------------------------------------
	//Method: Read
	//
	//Reads from data stream 1 (M&C) or 2 (high-speed data)
	//
	//If everything goes well, the number of bytes read (IN from device) will be returned.
	//Otherwise, an error code will be returned.
	//
	//Returns: number of bytes read (could be less than requested)
	// Blocking until either data is received or 100ms timeout.
	//
	//Error codes: 
	//-10 no handle to the driver is present. User needs to call openHandle() first.
	//-11 Missing a pointer to the buffer to hold the read data
	//-13 Can't read a negative number of bytes
	//-14 Could not detect M&C IN endpoint in descriptors
	//-15 Could not detect high-speed data IN endpoint in descriptors
	//-16 invalid dataStream input selection
	//-18 Some other error reading from USB device
	// -19 timeout reading
	//--------------------------------------------------------------------------------
	public native int Read(int dataStream, int noOfBytesToRead,
			byte[] bufferWithValuesToRead);

	/*******************************************************************************
	 *******************************************************************************/

	/** Sends a <code>String</code> over an existing USB connection to Data Stream 1
	 *
	 * @param messageToBeSent  a <code>String</code> to be sent over the
	 * existing USB connection to Data Stream 1, excluding the carriage return and the
	 * linefeed at the end
	 *
	 * Returns the number of bytes sent over USB (should be message length + 2).
	 */
	public int sendData(String messageToBeSent) {
		int status = 0;
		byte[] valuesToBeWritten = messageToBeSent.getBytes();
		byte[] finalValuesToBeWritten = new byte[messageToBeSent.length() + 2];
		byte[] valuesToBeWritten1 = new byte[] { (byte) 0x0D, //CR
				(byte) 0x0A }; //LF,
		System.arraycopy(valuesToBeWritten, 0, finalValuesToBeWritten, 0,
				valuesToBeWritten.length);
		System.arraycopy(valuesToBeWritten1, 0, finalValuesToBeWritten,
				valuesToBeWritten.length, valuesToBeWritten1.length);

		status = Write(1, finalValuesToBeWritten.length, finalValuesToBeWritten);
		return status;
	}

	/*******************************************************************************
	 *
	 *******************************************************************************/
	/** Sends a <code>String</code> over an existing USB connection to Data Stream 1
	 *  and waits for a reply
	 *
	 * @param messageToBeSent  a <code>String</code> to be sent over the
	 * existing USB connection to Data Stream 1, excluding the carriage return and the
	 * linefeed at the end
	 * @param timeOutValue specifies how many milliseconds after the
	 * message was sent the waiting for a reply will cease and Timeout will be
	 * returned
	 *
	 * @return  If a timeout occurs then the <code>String</code> Timeout will
	 * be returned. Otherwise, the received reply message, excluding
	 * the initial @ and the carriage return and linefeed at the end, will be
	 * returned.
	 */
	public String sendAndReceiveData(String messageToBeSent, int timeOutValue) {

		final boolean PRINT_OUT_DEBUG_MESSAGES = false;
		int status = 0;
		String responseMessage = "";
		byte[] valuesRead = new byte[256];
		boolean atFound = false;
		boolean endOfMessageDetected = false;

		status = sendData(messageToBeSent);
		if(status != (messageToBeSent.length() + 2)){
			// error sending.
			responseMessage = "USB tx error "+status + " != " + messageToBeSent.length();
			return responseMessage;
		}

		timeIsUp = false;
		startTimer(timeOutValue);	

		while ((!endOfMessageDetected) && (!timeIsUp)) {
			status = Read(1, valuesRead.length, valuesRead);
			for (int i = 0; i < status; i++) {
				if (status > 0) {
					if (atFound == true) {
						// First @ is found, now look for '\r' if not already found
						if ((valuesRead[i] == 0x0D) || (valuesRead[i] == 0x0A)) {
							// The current character is '\n' or '\r' and thus end of string
							endOfMessageDetected = true;
							timeoutTimerTask.cancel();
						} else {
							// The carriage return is not found; keep appending the characters
							responseMessage = responseMessage
									+ (char) valuesRead[i];
						}
					} else {
						// First @ is not found:
						if (valuesRead[i] == 0x40) {
							// Current character is @
							atFound = true;
						}
					}
				}
			}
		}
		if (PRINT_OUT_DEBUG_MESSAGES) {
			System.out.println("Received the following to the ComBlock: "
					+ responseMessage);
		}
		if ((endOfMessageDetected == false) && (timeIsUp)) {
			responseMessage = "Timeout";
			return responseMessage;
		}
/*
		// new 7/26/10 AZ
		// Make sure that there are no trailing messages waiting in a USB driver,
		// whether at tx or rx. Defensive code writing.
		// Otherwise logjam: each new query returns the previous answer.
		// WELL INTENTIONED BUT SLOWS DOWN RESPONSE WHEN THERE IS NOTHING ELSE TO READ

		status = 0;
		while (true) {
			try {
				status = Read(1, valuesRead.length, valuesRead);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				//e.printStackTrace();
			}
			if (status <= 0)
				break; // nothing to read
		}
*/
		return responseMessage;
	}
	
	// returns the number of bytes discarded. Or negative if error (see Read error codes)
	public int empty_rxqueue(){
		// attempt to read rx queue. Discard any data stuck in the queue
		// returns the number of bytes discarded
		byte[] valuesRead = new byte[256];
		int status = 0;

		status = Read(1, valuesRead.length, valuesRead);
		if(status == -19){	// timeout. nothing to read
			return 0;
		} else if(status < 0){
			return status;	// return the number of bytes discarded
		} else {
			return status;	// return the number of bytes discarded
		}
	}

	/*******************************************************************************
	 *
	 *******************************************************************************/
	/** Timer
	 *
	 * @param waitForResponseTime time to wait before timing out.
	 */
	private void startTimer(long waitForResponseTime) {
		timeoutTimerTask = new TimerTask() {
			public void run() {
				timeIsUp = true;
			}
		};
		Timer timeoutTimer = new Timer();
		timeoutTimer.schedule(timeoutTimerTask, waitForResponseTime);
	}
}
/*******************************************************************************
 *
 *******************************************************************************/
