// USB.java
// Java interface for USB Application
package comblock;

import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.util.Timer;
import java.util.TimerTask;
import org.usb4java.BufferUtils;
import org.usb4java.Device;
import org.usb4java.DeviceDescriptor;
import org.usb4java.DeviceHandle;
import org.usb4java.DeviceList;
import org.usb4java.LibUsb;
import org.usb4java.LibUsbException;
import org.usb4java.Transfer;
import org.usb4java.TransferCallback;

/**
 * <p>Title: USB </p>
 * <p>Description: Java Interface for USB Application</p>
 * <p>Copyright: Copyright (c) 2005-2019</p>
 * <p>Company: MSS</p>
 * @author Poornima Sarvadevabhatla / MSS
 * @author Alain Zarembowitch / MSS
 * @version 4.0
 * @since 9/7/19
 * 
 * Reference documents: 
 * - http://usb4java.org/
 * - https://github.com/usb4java/usb4java-examples/blob/master/src/main/java/org/usb4java/examples/AsyncBulkTransfer.java
 * 
 * Rev 3 1/9/11 AZ
 * Stop using the separate usb.jar 
 * Change driver to winusb for compatibility with newer Windows operating systems (7)
 * Use usbdll.dll (instead of the previous rev usbcpp.dll)
 * Removed exceptions. Errors are returned as negative numbers.
 * Made the native methods public... no need for another wrapper.
 * Added method to empty rx queue
 * 
 * Rev 4 9/7/19 AZ
 * Total rework using USB4JAVA / LIBUSB and WINUSB driver, 
 * Now compatible with 64-bit Java (only option with adoptopenjdk)
 */

public class USB {
    private final short VENDOR_ID = 0x0000;
    private final short PRODUCT_ID = 0x0004;
    private final long TIMEOUT = 100;	// The communication timeout in milliseconds
    
    private DeviceList deviceList;
    private DeviceHandle handle;
	private boolean timeIsUp = false;
	private TimerTask timeoutTimerTask = null;
	
	/***************************************************************************/
	public USB() {

		// the application should initialize LibUsb.init(null) once at the start
		// It should also LibUsb.exit(null) upon closing the application.
		// Example:
		// Upon starting the application:
	    // int result = LibUsb.init(null);
	    // if (result != LibUsb.SUCCESS)
	    // {
	    //    throw new LibUsbException("Unable to initialize libusb", result);
	    // }
	    // and upon closing the application:
	    //		LibUsb.exit(null);
	}
	

    /**
     * This is the event handling thread. libusb doesn't start threads by its
     * own so it is our own responsibility to give libusb time to handle the
     * events in our own thread.
     */
    class EventHandlingThread extends Thread
    {
        /** If thread should abort. */
        private volatile boolean abort;

        /**
         * Aborts the event handling thread.
         */
        public void abort()
        {
            this.abort = true;
        }

        @Override
        public void run()
        {
            while (!this.abort)
            {
                // Let libusb handle pending events. This blocks until events
                // have been handled, a hotplug callback has been deregistered
                // or the specified time of 0.5 seconds (Specified in
                // Microseconds) has passed.
                int result = LibUsb.handleEventsTimeout(null, 500000);
                if (result != LibUsb.SUCCESS)
                    throw new LibUsbException("Unable to handle events", result);
            }
        }
    }

	//--------------------------------------------------------------------------------
	//Method: nActiveDevices
	//This method detects how many ComBlocks are connected via USB and ready to communicate.
	//
	//Returns the number of ComBlocks directly connected to this PC via USB.
	//
	//--------------------------------------------------------------------------------
	public int nActiveDevices() {
		int nComBlocks = 0;
	    // Read the USB device list
	    deviceList = new DeviceList();
	    int result = LibUsb.getDeviceList(null, deviceList);
	    if (result < 0) throw new LibUsbException("Unable to get device list", result);

	    try
	    {
	        // Iterate over all devices and scan for the right ones with matching VENDOR_ID and PRODUCT_ID
	        for (Device device: deviceList)
	        {
	            DeviceDescriptor descriptor = new DeviceDescriptor();
	            result = LibUsb.getDeviceDescriptor(device, descriptor);
	            if (result != LibUsb.SUCCESS) throw new LibUsbException("Unable to read device descriptor", result);
	            if (descriptor.idVendor() == VENDOR_ID && descriptor.idProduct() == PRODUCT_ID) {
	            	nComBlocks++;
	            }
	        }
	    }
	    finally
	    {
	        // Ensure the allocated device list is freed
	        LibUsb.freeDeviceList(deviceList, true);
	    }
	    return nComBlocks;
	}
	
	//--------------------------------------------------------------------------------
	//Method: getUsbDevices
	//return the names of all relevant USB connections to ComBlocks
	//--------------------------------------------------------------------------------
	public String[] getUsbDevices() {

		int nd = nActiveDevices(); 
		String[] usbNames = new String[nd];

		for (int i = 0; i < nd; i++) {
			usbNames[i] = "comblock_usb_" + i;
		}
		return usbNames;
	}


	//--------------------------------------------------------------------------------
	//Method: openHandle
	//Get a device handle for the device at the specified index
	//The handle will be kept for all other methods until the close method is called.
	//The java application can only communicate with one driver at a time.
	//Attempt to open multiple handles will result in closing previous ones.
	//
	//This method does more than just creating a device handle and a winusb handle, 
	//it also scans the endpoint descriptor to find the correct endpoint locations for
	//the two monitoring and control pipes.
	//
	//Returns: 0 when everything went OK.
	//
	//Error codes: 
	//-1 error
	//--------------------------------------------------------------------------------
	public int openHandle(int device_index) {
		handle = new DeviceHandle();
		int index_cntr = 0;
        // Iterate over all devices and scan for the right ones with matching VENDOR_ID and PRODUCT_ID
		
	    // Read the USB device list
	    deviceList = new DeviceList();
	    int result = LibUsb.getDeviceList(null, deviceList);
	    if (result < 0) throw new LibUsbException("Unable to get device list", result);

	    for (Device device: deviceList)
        {
            DeviceDescriptor descriptor = new DeviceDescriptor();
            result = LibUsb.getDeviceDescriptor(device, descriptor);
            if (result != LibUsb.SUCCESS) throw new LibUsbException("Unable to read device descriptor", result);
            if ((descriptor.idVendor() == VENDOR_ID) && descriptor.idProduct() == PRODUCT_ID) {
            	if(index_cntr == device_index) {
            		// found the right one
            		result = LibUsb.open(device, handle);
            		if (result != LibUsb.SUCCESS) throw new LibUsbException("Unable to open USB device", result);
            		// When you want to communicate with an interface or with endpoints of this interface 
            		// then you have to claim it before using it and you have to release it when you are finished. 
            		// claim interface 0
            		result = LibUsb.claimInterface(handle, 0);
            		if (result != LibUsb.SUCCESS) throw new LibUsbException("Unable to claim interface 0", result);
            		return 0;
           	}
            	index_cntr++;
            }
        }
		
		return -1;
	}

	//--------------------------------------------------------------------------------
	//Method: closeHandle
	//Close the existing handle (libusb)
	//Returns: n/a
	//
	//Error codes: n/a
	//--------------------------------------------------------------------------------
	public void closeHandle() {
		int result;
		if(handle != null) {
			// release interface 0
			result = LibUsb.releaseInterface(handle, 0);
			if (result != LibUsb.SUCCESS) throw new LibUsbException("Unable to release interface 0", result);
	        // Close the device
	        LibUsb.close(handle);
		}
	}

	//--------------------------------------------------------------------------------
	//Method: Write
	//
	//Write to data stream 1 (M&C) or 2 (high-speed data)
	//
	//Returns the number of bytes actually sent (could be less if there are communication problems)
	// Blocking until either data is sent or 100ms timeout.
	//
	//Error codes: 
	//-10 no handle to the driver is present. User needs to call openHandle() first.
	//-13 Can't write a negative number of bytes
	//-16 invalid dataStream input selection
	//--------------------------------------------------------------------------------
	public int Write(int dataStream, int noOfBytesToWrite,
			byte[] bufferWithValuesToWrite) {
		byte endpoint_no;
		
		if(handle == null)
			return -10;
		if(noOfBytesToWrite < 0)
			return -13;
		if(dataStream == 1) {
			// M&C OUT endpoint
			endpoint_no = 0x02;
		} else if(dataStream == 2) {
			// high-speed data OUT endpoint
			// Note: the LPC17XX ARM processor declares endpoints 5 for data, but unused.
			// The COM-1400,1200,etc with USB2.0 PHY and FPGA SIE use endpoint 4
			endpoint_no = 0x04;
			//endpoint_no = 0x05;	// LPC17XX 
		} else {
			return -16; // invalid dataStream input selection
		}

		
		ByteBuffer buffer = ByteBuffer.allocateDirect(noOfBytesToWrite);
		buffer.put(bufferWithValuesToWrite, 0, noOfBytesToWrite);
		IntBuffer transferred = IntBuffer.allocate(1);
		int result = LibUsb.bulkTransfer(handle, endpoint_no, buffer, transferred, TIMEOUT); 
		if (result != LibUsb.SUCCESS) throw new LibUsbException("Bulk OUT transfer failed", result);
		//System.out.println(transferred.get() + " bytes sent");
		
		return transferred.get();
	}

	//--------------------------------------------------------------------------------
	//Method: Read
	//
	//Reads from data stream 1 (M&C) or 2 (high-speed data)
	//
	//If everything goes well, the number of bytes read (IN from device) will be returned.
	//Otherwise, an error code will be returned.
	//
	//Returns: number of bytes read (could be less than requested)
	// Blocking until either data is received or 100ms timeout.
	//
	//Error codes: 
	// -4 if the device has been disconnected
	// -7 timeout reading
	// -8 overflow if the device offered more data
	// -9 if the endpoint halted
	//-10 no handle to the driver is present. User needs to call openHandle() first.
	//-13 Can't read a negative number of bytes
	//-16 invalid dataStream input selection
	//-18 Some other error reading from USB device
	//--------------------------------------------------------------------------------
	public int Read(int dataStream, int noOfBytesToRead,
			byte[] bufferWithValuesToRead) {
		byte endpoint_no;

		if(handle == null)
			return -10;
		if(noOfBytesToRead < 0)
			return -13;
		if(dataStream == 1) {
			// M&C IN endpoint
			endpoint_no = -126; //0x82;
		} else if(dataStream == 2) {
			// high-speed data IN endpoint
			// Note: the LPC17XX ARM processor declares endpoints 5 for data, but unused.
			// The COM-1400,1200,etc with USB2.0 PHY and FPGA SIE use endpoint 4
			endpoint_no = -124; //0x84;
			//endpoint_no = -123; // 0x85;	LPC17XX 
		} else {
			return -16; // invalid dataStream input selection
		}
		ByteBuffer buffer = ByteBuffer.allocateDirect(noOfBytesToRead);
		IntBuffer transferred = IntBuffer.allocate(1);
		int result = LibUsb.bulkTransfer(handle, endpoint_no, buffer, transferred, TIMEOUT); 
		if (result != LibUsb.SUCCESS) return result;	// return LIBUSB error code
		
		buffer.get(bufferWithValuesToRead);
		return transferred.get();
	}

	/*******************************************************************************
	 *******************************************************************************/

	/** Sends a <code>String</code> over an existing USB connection to Data Stream 1
	 *
	 * @param messageToBeSent  a <code>String</code> to be sent over the
	 * existing USB connection to Data Stream 1, excluding the carriage return and the
	 * linefeed at the end
	 *
	 * Returns the number of bytes sent over USB (should be message length + 2).
	 */
	public int sendData(String messageToBeSent) {
		int status = 0;
		byte[] valuesToBeWritten = messageToBeSent.getBytes();
		byte[] finalValuesToBeWritten = new byte[messageToBeSent.length() + 2];
		byte[] valuesToBeWritten1 = new byte[] { (byte) 0x0D, //CR
				(byte) 0x0A }; //LF,
		System.arraycopy(valuesToBeWritten, 0, finalValuesToBeWritten, 0,
				valuesToBeWritten.length);
		System.arraycopy(valuesToBeWritten1, 0, finalValuesToBeWritten,
				valuesToBeWritten.length, valuesToBeWritten1.length);

		status = Write(1, finalValuesToBeWritten.length, finalValuesToBeWritten);
		return status;
	}

	/*******************************************************************************
	 *
	 *******************************************************************************/
	/** Sends a <code>String</code> over an existing USB connection to Data Stream 1
	 *  and waits for a reply
	 *
	 * @param messageToBeSent  a <code>String</code> to be sent over the
	 * existing USB connection to Data Stream 1, excluding the carriage return and the
	 * linefeed at the end
	 * @param timeOutValue specifies how many milliseconds after the
	 * message was sent the waiting for a reply will cease and Timeout will be
	 * returned
	 *
	 * @return  If a timeout occurs then the <code>String</code> Timeout will
	 * be returned. Otherwise, the received reply message, excluding
	 * the initial @ and the carriage return and linefeed at the end, will be
	 * returned.
	 */
	public String sendAndReceiveData(String messageToBeSent, int timeOutValue) {

		final boolean PRINT_OUT_DEBUG_MESSAGES = false;
		int status = 0;
		String responseMessage = "";
		byte[] valuesRead = new byte[256];
		boolean atFound = false;
		boolean endOfMessageDetected = false;

		status = sendData(messageToBeSent);
		if(status != (messageToBeSent.length() + 2)){
			// error sending.
			responseMessage = "USB tx error "+status + " != " + messageToBeSent.length();
			return responseMessage;
		}
		
//		TOO SLOW  10/13/11  not worth it// test test test new wait...
//		// the purpose of the delay is not to ask for read data before the ComBlock microcontroller has had time to prepare a response???
//		// typical.. 5 ms or so.
//		// 9/11/11
//		// Maybe wait a little after the transmission
//		if (ImportantConstants.WAIT_AFTER_USB_SENDS > 0) {
//			try {
//				Thread.sleep(ImportantConstants.WAIT_AFTER_USB_SENDS);
//			} catch (Exception e) {
//			}
//		}

		timeIsUp = false;
		startTimer(timeOutValue);	

		while ((!endOfMessageDetected) && (!timeIsUp)) {
			status = Read(1, valuesRead.length, valuesRead);
			for (int i = 0; i < status; i++) {
				if (status > 0) {
					if (atFound == true) {
						// First @ is found, now look for '\r' if not already found
						if ((valuesRead[i] == 0x0D) || (valuesRead[i] == 0x0A)) {
							// The current character is '\n' or '\r' and thus end of string
							endOfMessageDetected = true;
							timeoutTimerTask.cancel();
						} else {
							// The carriage return is not found; keep appending the characters
							responseMessage = responseMessage
									+ (char) valuesRead[i];
						}
					} else {
						// First @ is not found:
						if (valuesRead[i] == 0x40) {
							// Current character is @
							atFound = true;
						}
					}
				}
			}
		}
		if (PRINT_OUT_DEBUG_MESSAGES) {
			System.out.println("Received the following to the ComBlock: "
					+ responseMessage);
		}
		if ((endOfMessageDetected == false) && (timeIsUp)) {
			responseMessage = "Timeout";
			return responseMessage;
		}
/*
		// new 7/26/10 AZ
		// Make sure that there are no trailing messages waiting in a USB driver,
		// whether at tx or rx. Defensive code writing.
		// Otherwise logjam: each new query returns the previous answer.
		// WELL INTENTIONED BUT SLOWS DOWN RESPONSE WHEN THERE IS NOTHING ELSE TO READ

		status = 0;
		while (true) {
			try {
				status = Read(1, valuesRead.length, valuesRead);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				//e.printStackTrace();
			}
			if (status <= 0)
				break; // nothing to read
		}
*/
		return responseMessage;
	}
	
	// returns the number of bytes discarded. Or negative if error (see Read error codes)
	public int empty_rxqueue(){
		// attempt to read rx queue. Discard any data stuck in the queue
		// returns the number of bytes discarded
		byte[] valuesRead = new byte[256];
		int status = 0;

		status = Read(1, valuesRead.length, valuesRead);
		if(status == -19){	// timeout. nothing to read
			return 0;
		} else if(status < 0){
			return status;	// return the number of bytes discarded
		} else {
			return status;	// return the number of bytes discarded
		}
	}

	/*******************************************************************************
	 *
	 *******************************************************************************/
	/** Timer
	 *
	 * @param waitForResponseTime time to wait before timing out.
	 */
	private void startTimer(long waitForResponseTime) {
		timeoutTimerTask = new TimerTask() {
			public void run() {
				timeIsUp = true;
			}
		};
		Timer timeoutTimer = new Timer();
		timeoutTimer.schedule(timeoutTimerTask, waitForResponseTime);
	}
}
/*******************************************************************************
 *
 *******************************************************************************/
